from django.apps import apps
from django.utils.html import mark_safe
from mako.exceptions import RichTraceback
from mako.template import Template as MakoTemplate
from mako.runtime import Context as MakoContext, _populate_self_namespace

import io
import os, os.path


def template_inheritance(obj):
    '''
    Generator that iterates the template and its ancestors.
    The order is from most specialized (furthest descendant) to
    most general (furthest ancestor).

    obj can be either:
        1. Mako Template object
        2. Mako `self` object (available within a rendering template)
    '''
    if isinstance(obj, MakoTemplate):
        obj = create_mako_context(obj)['self']
    elif isinstance(obj, MakoContext):
        obj = obj['self']
    while obj is not None:
        yield obj.template
        obj = obj.inherits


def create_mako_context(template_obj, **kwargs):
    # I'm hacking into private Mako methods here, but I can't see another
    # way to do this.  Hopefully this can be rectified at some point.
    kwargs.pop('self', None)  # some contexts have self in them, and it messes up render_unicode below because we get two selfs
    runtime_context = MakoContext(io.StringIO(), **kwargs)
    runtime_context._set_with_template(template_obj)
    _, mako_context = _populate_self_namespace(runtime_context, template_obj)
    return mako_context


def get_template_debug(template_name, error):
    '''
    This structure is what Django wants when errors occur in templates.
    It gives the user a nice stack trace in the error page during debug.
    '''
    # This is taken from mako.exceptions.html_error_template(), which has an issue
    # in Py3 where files get loaded as bytes but `lines = src.split('\n')` below
    # splits with a string. Not sure if this is a bug or if I'm missing something,
    # but doing a custom debugging template allows a workaround as well as a custom
    # DMP look.
    # I used to have a file in the templates directory for this, but too many users
    # reported TemplateNotFound errors. This function is a bit of a hack, but it only
    # happens during development (and mako.exceptions does this same thing).
    #   /justification
    stacktrace_template = MakoTemplate(r"""
<%! from mako.exceptions import syntax_highlight, pygments_html_formatter %>
<style>
    .stacktrace { margin:5px 5px 5px 5px; }
    .highlight { padding:0px 10px 0px 10px; background-color:#9F9FDF; }
    .nonhighlight { padding:0px; background-color:#DFDFDF; }
    .sample { padding:10px; margin:10px 10px 10px 10px;
                font-family:monospace; }
    .sampleline { padding:0px 10px 0px 10px; }
    .sourceline { margin:5px 5px 10px 5px; font-family:monospace;}
    .location { font-size:80%; }
    .highlight { white-space:pre; }
    .sampleline { white-space:pre; }

    % if pygments_html_formatter:
        ${pygments_html_formatter.get_style_defs() | n}
        .linenos { min-width: 2.5em; text-align: right; }
        pre { margin: 0; }
        .syntax-highlighted { padding: 0 10px; }
        .syntax-highlightedtable { border-spacing: 1px; }
        .nonhighlight { border-top: 1px solid #DFDFDF;
                        border-bottom: 1px solid #DFDFDF; }
        .stacktrace .nonhighlight { margin: 5px 15px 10px; }
        .sourceline { margin: 0 0; font-family:monospace; }
        .code { background-color: #F8F8F8; width: 100%; }
        .error .code { background-color: #FFBDBD; }
        .error .syntax-highlighted { background-color: #FFBDBD; }
    % endif

    ## adjustments to Django css
    table.source {
        background-color: #fdfdfd;
    }
    table.source > tbody > tr > th {
        width: auto;
    }
    table.source > tbody > tr > td {
        font-family: inherit;
        white-space: normal;
        padding: 15px;
    }
    #template {
        background-color: #b3daff;
    }

</style>
<%

    src = tback.source
    line = tback.lineno
    if isinstance(src, bytes):
        src = src.decode()
    if src:
        lines = src.split('\n')
    else:
        lines = None
%>
<h3>${tback.errorname}: ${tback.message}</h3>

% if lines:
    <div class="sample">
    <div class="nonhighlight">
    % for index in range(max(0, line-4),min(len(lines), line+5)):
        <%
        if pygments_html_formatter:
            pygments_html_formatter.linenostart = index + 1
        %>
        % if index + 1 == line:
        <%
        if pygments_html_formatter:
            old_cssclass = pygments_html_formatter.cssclass
            pygments_html_formatter.cssclass = 'error ' + old_cssclass
        %>
            ${lines[index] | n,syntax_highlight(language='mako')}
        <%
        if pygments_html_formatter:
            pygments_html_formatter.cssclass = old_cssclass
        %>
        % else:
            ${lines[index] | n,syntax_highlight(language='mako')}
        % endif
    % endfor
    </div>
    </div>
% endif

<div class="stacktrace">
% for (filename, lineno, function, line) in tback.reverse_traceback:
    <div class="location">${filename}, line ${lineno}:</div>
    <div class="nonhighlight">
    <%
       if pygments_html_formatter:
           pygments_html_formatter.linenostart = lineno
    %>
        <div class="sourceline">${line | n,syntax_highlight(filename)}</div>
    </div>
% endfor
</div>
""")
    tback = RichTraceback(error, error.__traceback__)
    lines = stacktrace_template.render_unicode(tback=tback)
    return {
        'message': '',
        'source_lines': [
            ( '', mark_safe(lines) ),
        ],
        'before': '',
        'during': '',
        'after': '',
        'top': 0,
        'bottom': 0,
        'total': 0,
        'line': tback.lineno or 0,
        'name': template_name,
        'start': 0,
        'end': 0,
    }



import os, shutil

# this dict of options is merged with the current project's TEMPLATES entry for DMP
DEFAULT_OPTIONS = {
    # the default app and page to render in Mako when the url is too short
    # if None (no default app), DMP will not capture short URLs
    'DEFAULT_APP': 'homepage',
    'DEFAULT_PAGE': 'index',

    # functions to automatically add variables to the params/context before templates are rendered
    'CONTEXT_PROCESSORS': [
        'django.template.context_processors.static',            # adds "STATIC_URL" from settings.py
        'django.template.context_processors.debug',             # adds debug and sql_queries
        'django.template.context_processors.request',           # adds "request" object
        'django.contrib.auth.context_processors.auth',          # adds "user" and "perms" objects
        'django.contrib.messages.context_processors.messages',  # adds messages from the messages framework
        'django_mako_plus.context_processors.settings',         # adds "settings" dictionary
    ],

    # identifies where the Mako template cache will be stored, relative to each template directory
    'TEMPLATES_CACHE_DIR': '__dmpcache__',

    # the default encoding of template files
    'DEFAULT_TEMPLATE_ENCODING': 'utf-8',

    # imports for every template
    'DEFAULT_TEMPLATE_IMPORTS': [
        # alternative syntax blocks within your Mako templates
        # 'from django_mako_plus import django_syntax, jinja2_syntax, alternate_syntax',

        # the next two lines are just examples of including common imports in templates
        # 'from datetime import datetime',
        # 'import os, os.path, re, json',
    ],

    # whether autoescaping of expressions is on or off
    'AUTOESCAPE': True,

    # the converter class to use for parameter conversion
    # this should be ParameterConverter or a subclass of it
    'PARAMETER_CONVERTER': 'django_mako_plus.converter.ParameterConverter',

    # whether to send the custom DMP signals -- set to False for a slight speed-up in router processing
    # determines whether DMP will send its custom signals during the process
    'SIGNALS': False,

    # static file providers (see "static file" docs for full options here)
    'CONTENT_PROVIDERS': [
        # adds JS context - this should normally be listed FIRST
        { 'provider':   'django_mako_plus.JsContextProvider' },

        # Sass compiler and link generator
        # { 'provider':   'django_mako_plus.CompileScssProvider',
        #   'sourcepath': lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.scss'),
        #   'targetpath': lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.scss.css'),
        #   'command':    lambda p: [ shutil.which('sass'), f'--load-path="{BASE_DIR}"', p.sourcepath, p.targetpath ] },
        # { 'provider':   'django_mako_plus.CssLinkProvider',
        #   'filepath':   lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.scss.css') },

        # Less compiler and link generator
        # { 'provider':   'django_mako_plus.CompileLessProvider',
        #   'sourcepath': lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.less'),
        #   'targetpath': lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.less.css'),
        #   'command':    lambda p: [ shutil.which('lessc'), f'--source-map', p.sourcepath, p.targetpath ] },
        # { 'provider':   'django_mako_plus.CssLinkProvider',
        #   'filepath':   lambda p: os.path.join(p.app_config.name, 'styles', p.template_relpath + '.less.css') },

        # generic compiler and link generator (see DMP docs, same options as other entries here)
        # { 'provider':   'django_mako_plus.CompileProvider' },
        # { 'provider':   'django_mako_plus.LinkProvider' },

        # link generators for regular JS and CSS: app/scripts/*.js and app/styles/*.css
        { 'provider':   'django_mako_plus.CssLinkProvider' },
        { 'provider':   'django_mako_plus.JsLinkProvider' },

        # link generators for app/scripts/__bundle__.js (webpack bundler)
        # { 'provider':   'django_mako_plus.WebpackJsLinkProvider' },
    ],

    # webpack file discovery, used by `manage.py dmp_webpack` to generate __entry__.js files
    'WEBPACK_PROVIDERS': [
        # finders for app/scripts/*.js and app/styles/*.css
        { 'provider': 'django_mako_plus.JsLinkProvider' },
        { 'provider': 'django_mako_plus.CssLinkProvider' },
    ],

    # additional template dirs to search
    'TEMPLATES_DIRS': [
        # '/var/somewhere/templates/',
    ],
}



from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.forms.utils import flatatt
import logging
import os
from ..util import crc32
from ..util import log
from .base import BaseProvider


DUPLICATES_KEY = '_LinkProvider_Filename_Cache_'

#####################################################
###   LinkProvider abstract base class

class LinkProvider(BaseProvider):
    '''
    Renders links like <link> and <script> based on the name of the template
    and supertemplates.
    '''
    DEFAULT_OPTIONS = {
        # the path to search for, relative to the project directory.  possible values are:
        #   1. None: a default path is used, such as "{app}/{subdir}/{filename.ext}", prefixed
        #      with the static root at production; see subclasses for their default filenames.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: used directly
        'filepath': None,

        # the link to be inserted into the template. possible values:
        #   1. None: a default path is used, such as "<script src="..."></script>, prefixed
        #      with the static url at production; see subclasses for the default link format.
        #   2. function, lambda, or other callable: called as func(provider) and
        #      should return a string
        #   3. str: inserted directly into the template
        'link': None,

        # extra attributes for the link element
        'link_attrs': {},

        # if a template is rendered more than once in a request, should we link more than once?
        # defaults are: css=False, js=True, bundled_js=False
        'skip_duplicates': False,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.filepath, self.absfilepath, self.mtime, self.version_id = self.get_cache_item()

        except AttributeError:
            self.filepath = self.build_source_filepath()
            self.absfilepath = os.path.join(
                settings.BASE_DIR if settings.DEBUG else settings.STATIC_ROOT,
                self.filepath,
            )
            # file time and version hash
            try:
                self.mtime = int(os.stat(self.absfilepath).st_mtime)
                # version_id combines file modification time and the CRC32 checksum of file bytes
                self.version_id = (self.mtime << 32) | crc32(self.absfilepath)
            except FileNotFoundError:
                self.mtime = 0
                self.version_id = 0
            self.set_cache_item((self.filepath, self.absfilepath, self.mtime, self.version_id))

        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created for %s: [%s]', repr(self), self.filepath, 'will link' if self.mtime > 0 else 'will skip nonexistent file')


    ### Source Filepath Building Methods ###

    def build_source_filepath(self):
        # if defined in settings, run the function or return the string
        if self.options['filepath'] is not None:
            return self.options['filepath'](self) if callable(self.options['filepath']) else self.options['filepath']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `filepath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_filepath()

    def build_default_filepath(self):
        # this method is overridden in CssLinkProvider and JsLinkProvider lower in this file
        raise ImproperlyConfigured('{} must set `filepath` in options (or a subclass can override build_default_filepath).'.format(self.__class__.__qualname__))


    ### Target Link Building Methods ###

    def build_target_link(self):
        # if defined in settings, run the function or return the string
        if self.options['link'] is not None:
            return self.options['link'](self) if callable(self.options['link']) else self.options['link']
        # build the default
        if self.app_config is None:
            log.warn('{} skipped: template %s not in project subdir and `sourcepath` not in settings', (self.__class__.__qualname__, self.template_relpath))
        return self.build_default_link()

    def build_default_link(self):
        # this method is overridden in CssLinkProvider and JsLinkProvider lower in this file
        raise ImproperlyConfigured('{} must set `link` in options (or a subclass can override build_default_link).'.format(self.__class__.__qualname__))


    ### Provider Run Methods ###

    def get_already_generated(self):
        # try to cache in the request, but if request is None, use the provider run
        # note that the single key string skips duplicates across different instances, not just within this instance
        placeholder = self.provider_run.request.dmp if self.provider_run.request is not None else self.provider_run
        try:
            return getattr(placeholder, DUPLICATES_KEY)
        except AttributeError:
            cache = set()
            setattr(placeholder, DUPLICATES_KEY, cache)
            return cache

    def provide(self):
        # short circuit if the file doesn't exist
        if self.mtime == 0:
            return
        # short circut if we're skipping duplicates and we've already seen this one
        if self.options['skip_duplicates']:
            already_generated = self.get_already_generated()
            if self.absfilepath in already_generated:
                if log.isEnabledFor(logging.DEBUG):
                    log.debug('%s skipped duplicate %s', repr(self), self.filepath)
                return
            already_generated.add(self.absfilepath)

        # if we get here, generate the link
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s linking %s', repr(self), self.filepath)
        self.write(self.build_target_link())


#####################################
###  CssLinkProvider

class CssLinkProvider(LinkProvider):
    '''Generates a CSS <link>'''
    DEFAULT_OPTIONS = {
        'group': 'styles',
        # if a template is rendered more than once in a request, we usually don't
        # need to include the css again.
        'skip_duplicates': True,
    }

    def build_default_filepath(self):
        '''Called when 'filepath' is not defined in the settings'''
        return os.path.join(
            self.app_config.name,
            'styles',
            self.template_relpath + '.css',
        )

    def build_default_link(self):
        '''Called when 'link' is not defined in the settings'''
        attrs = {}
        attrs["rel"] = "stylesheet"
        attrs["href"] ="{}?{:x}".format(
            os.path.join(settings.STATIC_URL, self.filepath).replace(os.path.sep, '/'),
            self.version_id,
        )
        attrs.update(self.options['link_attrs'])
        attrs["data-context"] = self.provider_run.uid       # can't be overridden
        return '<link{} />'.format(flatatt(attrs))


##############################################
###   JsLinkProvider

class JsLinkProvider(LinkProvider):
    '''Generates a JS <script>.'''
    DEFAULT_OPTIONS = {
        'group': 'scripts',
        # if a template is rendered more than once in a request, we should link each one
        # so the script runs again each time the template runs
        'skip_duplicates': False,
    }

    def build_default_filepath(self):
        '''Called when 'filepath' is not defined in the settings'''
        return os.path.join(
            self.app_config.name,
            'scripts',
            self.template_relpath + '.js',
        )

    def build_default_link(self):
        '''Called when 'link' is not defined in the settings'''
        attrs = {}
        attrs["src"] = "{}?{:x}".format(
            os.path.join(settings.STATIC_URL, self.filepath).replace(os.path.sep, '/'),
            self.version_id,
        )
        attrs.update(self.options['link_attrs'])
        attrs["data-context"] = self.provider_run.uid       # can't be overridden
        return '<script{}></script>'.format(flatatt(attrs))



from django.utils.module_loading import import_string
import json
import logging
from ..version import __version__
from ..util import log
from .base import BaseProvider

###################################
###  JS Context Provider

class JsContextProvider(BaseProvider):
    '''
    Adds all js_context() variables to DMP_CONTEXT.
    '''
    DEFAULT_OPTIONS = {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'scripts',
        # the encoder to use for the JSON structure
        'encoder': 'django.core.serializers.json.DjangoJSONEncoder',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.encoder = import_string(self.options['encoder'])
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s created', repr(self))

    def provide(self):
        # we output on the first run through - the context is only needed once
        if not self.is_first():
            return

        # generate the context dictionary
        data = {
            'id': self.provider_run.uid,
            'version': __version__,
            'templates': [ '{}/{}'.format(p.app_config.name, p.template_relpath) for p in self.iter_related() ],
            'app': self.provider_run.request.dmp.app if self.provider_run.request is not None else None,
            'page': self.provider_run.request.dmp.page if self.provider_run.request is not None else None,
            'log': log.isEnabledFor(logging.DEBUG),
            'values': {
                'id': self.provider_run.uid,
            },
        }
        for k in self.provider_run.context.keys():
            if isinstance(k, jscontext):
                value = self.provider_run.context[k]
                data['values'][k] = value.__jscontext__() if callable(getattr(value, '__jscontext__', None)) else value

        # output the script
        self.write('<script>')
        self.write('DMP_CONTEXT.set({data});'.format(
            data=json.dumps(data, cls=self.encoder, separators=(',', ':'))
        ))
        self.write('</script>')


class jscontext(str):
    '''
    Marks a key in the context dictionary as a JS context item.
    JS context items are sent to the template like normal,
    but they are also added to the runtime JS namespace.

    See the tutorial for more information on this function.
    '''
    # no code needed, just using the class for identity



from django import template
from django.apps import apps
from django.utils.safestring import mark_safe


###############################################################
###  DJANGO template tag to include a Mako template
###
###  This file is called "django_mako_plus.py" because it's
###  the convention for creating Django template tags.
###
###  See also django_mako_plus/filters.py

register = template.Library()

@register.simple_tag(takes_context=True)
def dmp_include(context, template_name, def_name=None, **kwargs):
    '''
    Includes a DMP (Mako) template into a normal django template.

    context:        automatically provided
    template_name:  specified as "app/template"
    def_name:       optional block to render within the template

    Example:
        {% load django_mako_plus %}
        {% dmp_include "homepage/bsnav_dj.html" %}
        or
        {% dmp_include "homepage/bsnav_dj.html" "blockname" %}
    '''
    dmp = apps.get_app_config('django_mako_plus')
    template = dmp.engine.get_template(template_name)
    dmpcontext = context.flatten()
    dmpcontext.update(kwargs)
    return mark_safe(template.render(
        context=dmpcontext,
        request=context.get('request'),
        def_name=def_name
    ))



from django.template import engines
from django.template import TemplateDoesNotExist
from mako.runtime import supports_caller

###
###  Mako-style tags that DMP provides
###


###############################################################
###  Include Django templates
###

def django_include(context, template_name, **kwargs):
    '''
    Mako tag to include a Django template withing the current DMP (Mako) template.
    Since this is a Django template, it is search for using the Django search
    algorithm (instead of the DMP app-based concept).
    See https://docs.djangoproject.com/en/2.1/topics/templates/.

    The current context is sent to the included template, which makes all context
    variables available to the Django template. Any additional kwargs are added
    to the context.
    '''
    try:
        djengine = engines['django']
    except KeyError as e:
        raise TemplateDoesNotExist("Django template engine not configured in settings, so template cannot be found: {}".format(template_name)) from e
    djtemplate = djengine.get_template(template_name)
    djcontext = {}
    djcontext.update(context)
    djcontext.update(kwargs)
    return djtemplate.render(djcontext, context['request'])



#########################################################
###  Template autoescaping on/off

# attaching to `caller_stack` because it's the same object
# throughout rendering of a template inheritance
AUTOESCAPE_KEY = '__dmp_autoescape'

def is_autoescape(context):
    return bool(getattr(context.caller_stack, AUTOESCAPE_KEY, True))


def _toggle_autoescape(context, escape_on=True):
    '''
    Internal method to toggle autoescaping on or off. This function
    needs access to the caller, so the calling method must be
    decorated with @supports_caller.
    '''
    previous = is_autoescape(context)
    setattr(context.caller_stack, AUTOESCAPE_KEY, escape_on)
    try:
        context['caller'].body()
    finally:
        setattr(context.caller_stack, AUTOESCAPE_KEY, previous)


@supports_caller
def autoescape_on(context):
    '''
    Mako tag to enable autoescaping for a given block within a template,
    (individual filters can still override with ${ somevar | n }).

    Example:
        <%namespace name="dmp" module="django_mako_plus.tags"/>
        <%dmp:autoescape_on>
            ${ somevar } will be autoescaped.
        </%dmp:autoescape_on>
    '''
    _toggle_autoescape(context, True)
    return ''


@supports_caller
def autoescape_off(context):
    '''
    Mako tag to disable autoescaping for a given block within a template,
    (individual filters can still override with ${ somevar | h }).

    Example:
        <%namespace name="dmp" module="django_mako_plus.tags"/>
        <%dmp:autoescape>
            ${ somevar } will not be autoescaped.
        </%dmp:autoescape>
    '''
    _toggle_autoescape(context, False)
    return ''
